package dev.pfaff.altargui.shader;

import dev.pfaff.glass.opengl.enums.GLVertexAttribPointerType;
import dev.pfaff.glass.opengl.enums.IGLVertexAttribPointerType;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.MemoryLayout;
import java.util.Objects;

import static java.lang.foreign.MemoryLayout.sequenceLayout;
import static java.lang.foreign.ValueLayout.JAVA_BYTE;

public sealed interface VertexComponentSpec {
	String name();

	MemoryLayout layout();

	record Padding(int byteSize) implements VertexComponentSpec {
		@Override
		public String name() {
			return "padding";
		}

		@Override
		public MemoryLayout layout() {
			return sequenceLayout(this.byteSize, JAVA_BYTE);
		}
	}

	record Attrib(String name, IGLVertexAttribPointerType type, boolean normalized, int width, @Nullable MemoryLayout customLayout) implements VertexComponentSpec {
		public Attrib {
			if (normalized && !(type instanceof GLVertexAttribPointerType)) {
				throw new UnsupportedOperationException("Only GLVertexAttribPointerType can be normalized");
			}
			if (customLayout != null) {
				checkLayoutCompatibility(type, type.layout(), customLayout);
			}
		}

		public Attrib(String name, IGLVertexAttribPointerType type, boolean normalized, int width) {
			this(name, type, normalized, width, null);
		}

		public Attrib(String name, IGLVertexAttribPointerType type, int width) {
			this(name, type, false, width);
		}

		@Override
		public MemoryLayout layout() {
			return this.customLayout != null ? this.customLayout : sequenceLayout(this.width, this.type.layout());
		}

		public Attrib withNormalized() {
			return new Attrib(name, type, true, width, customLayout);
		}

		public Attrib withCustomLayout(MemoryLayout layout) {
			Objects.requireNonNull(layout);
			if (customLayout != null) {
				checkLayoutCompatibility(type, customLayout, layout);
			}
			return new Attrib(name, type, normalized, width, layout);
		}

		private static void checkLayoutCompatibility(IGLVertexAttribPointerType type,
													 MemoryLayout base,
													 MemoryLayout custom) {
			if (custom.byteSize() != base.byteSize() || custom.byteAlignment() < base.byteAlignment()) {
				throw new IllegalArgumentException(STR."""
				Incompatible custom layout for \{type} attribute:
					Base layout: \{base}
					Custom layout: \{custom}
				""");
			}
		}
	}
}
